(thought
  :title "a moderately cursed dollcode implementation"
  :date "2024-10-24 23:18:14"
  :tags ["hacks" "programming" "python"]
  :description "implementing a dollcode translator in python, very, very badly."
  :content `(
    (p #[[
      recently, dollcode has become a thing frquently utilized by both this one's friends and other beings it observes.
      this is inconvenient, because many of them have written their own transcoders for the format, which can handle arbitrarily large numbers.
      at time of writing, the only two translators it has found (]] ~(link "https://noe.sh/dollcode") #[[, ]] ~(link "https://999eagle.moe/dollcode") #[[)
      do not support nubbers larger than 64 bit integers. thusly, it can read few of its friends encoded strings.]])
    
    (p #[[
      the optimal solution to such a problem is probably either asking the sender of a given message what said message says, or asking them for their transcoder.
      this one being this one, it did neither: below, one may find a very good and very reasonable dollcode transcoder. one can tell it is good and reasonable 
      because it contains the y combinator written in python lambda statements.]])
    
    (figure
      (figcaption "~/.local/bin/dollcode")
      (code
        ~(run "syntax-hl data/thoughts/dollcode/dollcode.py")))
    
    (p #[[
      python is a very good programming language. this is observable in that it allows integers to have up to 4300 digits]] 
       ~(run "make-footnote 'or, up to a value specified via sys.set_int_max_str_digits(some_int)'"))
    
    (p #[[one may wonder: "why is the shebang like that?" the answer to such a question is that it did not want to set the recursion limit inside the actual program
      (it did not realize it would hit recursion limit issues until after it was done writing the program), and it decided it would be amusing to have a really weird shebang line.]])
    
    (p #[[bafflingly, this program works:]])
    (code #[[
      ~ λ dollcode -a nat<br>
      ▖▖▖▖▘▖▖▖▖▘▌▖▖▖▘]])
    
    (p #[[it is very efficient:]])
    (code (pre
      ~(run "cat data/thoughts/dollcode/dollcode-invocation-memory-log")))
    (p #[[the reason for this is interesting: 
      looking at the the argument passed to the encode function, d, we can see it is an integer such that if the integer were to be represented in base 16 
      and split in to groups of two characters, each set of two characters would represent the byte that a character to be encoded is (ex: "nat" -> 0x6e6174 -> 6e, 61, 74 -> n, a, t).]])

    (p #[[
      after this process of encoding the string to be encoded as a potentially very very big integer, d defines two important internal functions: 
      we will call them a and b. a takes an argument n, and returns the result of b called on both return values of divmod(n-1, 3) (those being the quotient and modulus of n by 3). 
      the second function, b, takes two arguments: x, and y. 
      it then returns str(f(x)) + str(y), where f is defined by the y combinator to be the first function, a.]])

    (p #[[
      since, before b can return, a((n-1)//3) must first return, which in turn calls b with different arguments, we fill the stack with approximately log3(n)
      frames containing numbers smaller than the initial (arbitrarily large) number. not accounting for the overhead of the memory taken by each function, the
      number of bytes it takes to store an array of [n, n//3, n//9... 1] is approximately 2.7*n^2. if we plug 65536 in to our equation, we get
      2.7*65536^2 = 11596411699.2. dividing this by 1024^3, we get 10.8, which is very nearly the amount of memory, in gigabytes, this program uses to encode 65536 chars]])
    
    (p #[[it is also very fast:]] ~(run "make-footnote 'it is only like this for large inputs, it is near instant for short strings. the python interpreter really does not like this code.'"))
    (code (pre
      ~(run "cat data/thoughts/dollcode/dollcode-invocation-time")))
    
    (footnote "this post was rewritten on 2025-05-12, as part of the migration to the new website. the wording has been updated, however it remains similar to the original.")
    (br)))
